'''Game main module.

Contains the entry point used by the run_game.py script.

Feel free to put all your game code here, or in other modules in this "gamelib"
package.
'''
import gamelib
from gamelib import data, render, world, input, timeline, audio, scenegraph
#from SDL import *
import SDL
class core(object):
    
    def __init__(self, screen_size=(1024,768), fullscreen=False):
        SDL.SDL_Init(SDL.SDL_INIT_TIMER)
        self.screen = render.screen(width=screen_size[0], height=screen_size[1], fullscreen=fullscreen)
        self.screen_size = screen_size
        self.rcontext = render.render_context()
        self.input = input.Input(screen_size=screen_size)
        self.act = 'menu'
        #self.act= ''
        self.world = world.menu()
        #self.world = world.World(screen_size=screen_size, world_name=self.act)
        
        self.acontext = audio.audio_context()
        self.acontext.play_music('Dirt Rhodes.ogg')
        
        self.x_pos = 0.0
        self.y_pos = 0.0

        self.timeline = timeline.Timeline()
        self.timeline.add_action(time=2000.0, callback=self.cb_test, relative=timeline.time_now)

        self.fpsaim = 30.0

    def cb_test(self, thing):
        pass
        #print 'this is a callback from the timeline!'
        
        
    def main_loop(self):
        
        if self.world.end_act:
            self.world.end()
            if self.act == 'menu':
                self.act = 'actI'
            elif self.act == 'actI':
                self.act = 'actII'
            elif self.act == 'actII':
                self.act = 'actIII'
            elif self.act == 'actIII':
                self.act = 'actIV'
                
            self.world = world.World(screen_size=self.screen_size, world_name=self.act)
            self.world.end_act = False
        
        
        pre_ticks = SDL.SDL_GetTicks()
        
        self.timeline.step()
        self.input.pump()
        if self.input.do_left:
            self.x_pos -= 10.0
        elif self.input.do_right:
            self.x_pos += 10.0
        
        if self.input.do_down:
            self.y_pos -= 10.0
        elif self.input.do_up:
            self.y_pos += 10.0
            
        #print self.input.mouse_x, self.input.mouse_y
        
        
        
        self.rcontext.reset()
        #self.rcontext.set_screen_pos(self.x_pos, self.y_pos)
        #self.rcontext.set_screen()
        
        self.rcontext.set_blend('modulate')
        
        self.world.tick_world()

        self.world.draw_world()
        
        post_ticks = SDL.SDL_GetTicks()
        
        aim_ticks = int(1000 / self.fpsaim)

        self.screen.end_frame()
        
        if (post_ticks - pre_ticks) > aim_ticks:
            pass
            #print 'Running too slow :('
        else:
            SDL.SDL_Delay(aim_ticks - (post_ticks - pre_ticks))
        
    def run(self):
        do_quit = False
        while not do_quit and not self.input.do_quit:
            self.main_loop()
            
        
        
def usage():
    print """
HOW TO PLAY THE GAME:

Click on items of interest, try and divert your life towards somewhere happy.

COMMAND LINE SWITCHES:
--fullscreen: switches on fullscreen mode
--resolution=800x600: lets you specify a resolution to use

"""
import os, sys
def main(*args):
    
    #try:
        #import psyco
        #psyco.full()
        #psyco.cannotcompile(scenegraph)
    #except ImportError:
        #print 'Psyco does not work on this machine or it is not installed'
        #pass
    
    import getopt


    try:
        opts, args = getopt.getopt(sys.argv[1:], "hf", ["help", "fullscreen", "resolution=", "widescreen"])
    except getopt.GetoptError, err:
        # print help information and exit:
        print str(err) # will print something like "option -a not recognized"
        usage()
        sys.exit(2)

    fullscreen = False
    resolution = (1024, 768)
    level = 0

    for o, a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit()

        elif o in ("-f", "--fullscreen"):
            fullscreen = False

        elif o in ("--resolution"):
            try:
                resolution = a.split('x')
                resolution = (int(resolution[0]), int(resolution[1]))
            except:
                print 'sorry,', a, 'is not a valid string, resolution should be in the form of --resolution=800x600 or something'
                sys.exit()

        else:
            assert False, "unhandled option"
    
    maincore = core(screen_size=resolution, fullscreen=fullscreen)
    maincore.run()
    
    SDL.SDL_Quit()
