#!/usr/bin/env python
#
#       render.py
#       
#       Copyright 2009 gord <gordallott@gmail.com>
#       
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 3 of the License, or
#       (at your option) any later version.
#       
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#       
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
import os

from SDL import *
from SDL.image import *
import OpenGL
OpenGL.ERROR_CHECKING = False
from OpenGL.GL import *
from OpenGL.GLU import *

import numpy
from OpenGL.arrays import numpymodule
numpymodule.NumpyHandler.ERROR_ON_COPY = False

#shader_support = False
#try:
    #from OpenGL.GL.ARB.shader_objects import *
    #from OpenGL.GL.ARB.fragment_shader import *
    #from OpenGL.GL.ARB.vertex_shader import *
    #shader_support = True
#except:
    #print 'Shaders are not supported with your graphics chip'
    #shader_support = False    
import sys
try:
    from OpenGL.GL.ARB.vertex_buffer_object import *
except:
    print 'VertexBufferObject extension not present on this gpu, this game will *not* run. sorry :('
    sys.exit(-1)
    
import re

import gamelib.data as data
import gamelib.data
import math, random

# Found this on the internet somewhere, active state cookbook maybe?
# it might of been a wikipage i found for python decorators
# basically it just lets me set the set_attr and get_attr for variables quickly
def propertyd(function):
    keys = 'fget', 'fset', 'fdel'
    func_locals = {'doc':function.__doc__}
    def probe_func(frame, event, arg):
        if event == 'return':
            locals = frame.f_locals
            func_locals.update(dict((k, locals.get(k)) for k in keys))
            sys.settrace(None)
        return probe_func
    sys.settrace(probe_func)
    function()
    return property(**func_locals)

target_surface = None

class screen(object):
    
    def __init__(self, width=1024, height=640, fullscreen=False):
        #starts an sdl/opengl screen
        
        SDL_InitSubSystem(SDL_INIT_VIDEO)
        SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1)
        
        global target_surface
        self.screen = SDL_SetVideoMode(width, height, 8, SDL_OPENGL)
        SDL_WM_SetCaption('CSL-II', 'CSL-II')
        if fullscreen:
            print 'Fullscreen worked?', SDL_WM_ToggleFullScreen(self.screen)
        
        rmask, gmask, bmask, amask = (0, 0, 0, 0)
        if SDL_BYTEORDER == SDL_BIG_ENDIAN:
            rmask = 0xff000000
            gmask = 0x00ff0000
            bmask = 0x0000ff00
            amask = 0x000000ff
        else:
            rmask = 0x000000ff
            gmask = 0x0000ff00
            bmask = 0x00ff0000
            amask = 0xff000000

        target_surface = SDL_CreateRGBSurface(SDL_SWSURFACE, 1, 1, 32,
                                       rmask, gmask, bmask, amask)

        
        
        screen_size = (width, (width / 16) * 9)
        bar_size = (height - screen_size[1]) / 2
        
        #print screen_size
        #print bar_size
        
        glViewport(0, bar_size, width, screen_size[1])
        #glViewport(0, 0, 400, 300)
        
        self.width = width
        self.height = height
        self.aspect = float(width) / height
        self.view_size = (400, 224)
        
        self.context = render_context(view_size=self.view_size)
        self.context.set_projection()
        glClearColor(0.0, 0.0, 0.0, 1.0)
        

        
    def end_frame(self):
        
        SDL_GL_SwapBuffers()
        glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT)
        
        
class render_context(object):
    """ this class takes a lot from the CSL-I entry, there was some
    fairly nice code todo with translating and such that i like to keep
    
    this class is rather clever in that it can be instanced as many
    times as you like but it still sort of acts like there is only
    one version of this class, which allows for clever optimisations
    
    does this by accessing the .__class__. for each variable 
    which means that all the instances access the same data
    """
        
    view_size = (20, 20)
    base_scalar = (1.0, 1.0, 1.0)
    base_rotation = (0.0, 0.0, 0.0)
    base_position = (0.0, 0.0, -1.0)
    position = (0.0, 0.0, -0.0)
    rotation = (0.0, 0.0, 0.0)
    scalar = (1.0, 1.0)
    texture_scalar = (1.0, 1.0)
    
    blend_func = None
    
    vbo_vert = None
    vbo_color = None
    vbo_texcoord = None
    
    state_vert_array = False
    state_color_array = False
    state_texcoord_array = False
    
    tex_01 = None
    tex_02 = None
    tex_03 = None
    
    textures = {}
    current_texture = 0
    
    color = (1.0, 1.0, 1.0, 1.0)
        
    def __init__(self, view_size=None):
        if view_size:
            self.set_view_size(view_size)
            
    def set_view_size(self, view_size):
        self.__class__.view_size = view_size

    # These functions are basically copied stright from CSL-I, modified
    # to use the __class__ lookup method. they worked nice enough in CSL-I

    def set_projection(self):
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        
        glOrtho(-self.__class__.view_size[0] * 0.5, self.__class__.view_size[0] *0.5,
                -self.__class__.view_size[1] * 0.5, self.__class__.view_size[1] *0.5,
                -22.0, 22.0)
                
        glMatrixMode(GL_MODELVIEW)
        self.reset()
        
    def _reset_rotation(self):
        """ resets the rotation to 0 """
        if not self.__class__.rotation == (0.0, 0.0, 0.0): 
            glRotate(self.__class__.rotation[2],  0.0,  0.0, -1.0)
            glRotate(self.__class__.rotation[1],  0.0, -1.0,  0.0)
            glRotate(self.__class__.rotation[0], -1.0,  0.0,  0.0)

            self.__class__.rotation = (0.0, 0.0, 0.0)

    def _reset_scale(self):
        """ resets the scale to 1.0 """
        if not self.__class__.scalar == (1.0, 1.0):
            glScale(1.0/self.__class__.scalar[0], 1.0/self.__class__.scalar[1], 1.0)
            self.__class__.scalar = (1.0, 1.0)
            
    def _reset_texture_scale(self):
        """ resets the scale to 1.0 """
        if not self.__class__.texture_scalar == (1.0, 1.0):
            glScale(1.0/self.__class__.texture_scalar[0], 1.0/self.__class__.texture_scalar[1], 1.0)
            self.__class__.texture_scalar = (1.0, 1.0)

    def _reset_position(self):
        """ resets the translation to 0 """
        if not self.__class__.position == (0.0, 0.0, 0.0):
            glTranslate(-self.__class__.position[0], -self.__class__.position[1], -self.__class__.position[2])

            self.__class__.position = (0.0, 0.0, 0.0)

    def reset(self):
        self._reset_scale()
        self._reset_rotation()
        self._reset_position()

        glLoadIdentity()
        
        glMatrixMode(GL_TEXTURE)
        self._reset_texture_scale()
        glLoadIdentity()
        glMatrixMode(GL_MODELVIEW)

    def rotate_to_simple(self, value):
        """ rotates to the specified degrees """
        cx, cy, cz = self.__class__.rotation

        #glRotate(x - cx, 1.0, 0.0, 0.0)
        #glRotate(y - cy, 0.0, 1.0, 0.0)
        glRotate(value - cz, 0.0, 0.0, 1.0)

        self.__class__.rotation = (cx, cy, value)
        
    def rotate_to(self, x, y, z):
        """ rotate to the specified degrees """

        cx, cy, cz = self.__class__.rotation

        glRotate(x - cx, 1.0, 0.0, 0.0)
        glRotate(y - cy, 0.0, 1.0, 0.0)
        glRotate(z - cz, 0.0, 0.0, 1.0)

        self.__class__.rotation = (x, y, z)

    def scale_to(self, x, y):
        """ scale to the specified scalar """
        x = float(x)
        y = float(y)

        self._reset_scale()

        glScale(x, y, 1.0)

        self.__class__.scalar = (x, y)
        
    def texture_scale_to(self, x, y):
        """ scale to the specified texture scalar """
        x = float(x)
        y = float(y)
        
        glMatrixMode(GL_TEXTURE) #/* enter texture matrix mode */
        self._reset_texture_scale()
        glScale(x, y, 1.0)
        glMatrixMode(GL_MODELVIEW) #/* back to modelview mode */

    def move_to(self, x, y, z=-1):
        """ translate to the specified position """
        self._reset_scale()
        self._reset_rotation()

        cx, cy, cz = self.__class__.position
        glTranslate(x - cx, y - cy, z - cz)
        self.__class__.position = (x, y, z)
        
    def set_screen_zoom(self, zoom):
        self.__class__.base_scalar = (zoom, zoom, 1.0)

    def set_screen_pos(self, x, y):
        self.__class__.base_position = (x, y, 0.0)

    def set_screen_rot(self, angle):
        self.__class__.base_rotation = (0.0, 0.0, angle) 

    def set_screen(self):
        """ should be called once per frame to set the screen zoom/pos/rot"""
        glScale(self.__class__.base_scalar[0], self.__class__.base_scalar[1], self.__class__.base_scalar[2])
        glRotate(self.__class__.base_rotation[2], 0.0, 0.0, 1.0)
        glTranslate(-self.__class__.base_position[0], self.__class__.base_position[1], self.__class__.base_position[2])

        
    def select_texture(self, unit):

        if self.__class__.current_texture == unit:
            return

        if unit == 0:
            glActiveTexture(GL_TEXTURE0)

        if unit == 1:
            glActiveTexture(GL_TEXTURE1)

        if unit == 2:
            glActiveTexture(GL_TEXTURE2)

        if unit > 2 or unit < 0:
            return

        self.__class__.current_texture = unit

    def unbind_textures(self):

        if self.__class__.tex_03:
            self.select_texture(2)
            glBindTexture(GL_TEXTURE_2D, 0);
            self.__class__.tex_03 = None

        if self.__class__.tex_02:
            self.select_texture(1)
            glBindTexture(GL_TEXTURE_2D, 0);
            self.__class__.tex_02 = None

        if self.__class__.tex_01:
            self.select_texture(0)
            glBindTexture(GL_TEXTURE_2D, 0);
            self.__class__.tex_01 = None

    def bind_texture(self, *textures):

        self.unbind_textures()

        for texnum in reversed(range(0, len(textures))):
            self.select_texture(texnum)
            glBindTexture(GL_TEXTURE_2D, textures[texnum].id)

            if texnum == 0:
                self.__class__.tex_01 = textures[texnum]

            if texnum == 1:
                self.__class__.tex_02 = textures[texnum]

            if texnum == 2:
                self.__class__.tex_03 = textures[texnum]

# this is straight from CSL-I at the moment, needs to be re-written
    def create_texture(self, filename, **kwargs):
        
        #print filename, kwargs
        filepath = data.filepath(filename)
    
        if not os.path.exists(filepath):
            raise Exception, 'path %s does not exist' % (filepath)

        if self.__class__.textures.has_key(filepath):
            texture = self.__class__.textures[filepath]
        else:
            
            if not kwargs.has_key('ignore_anim') and 'anim' in filepath: 
                texture = Animation(filename, **kwargs)
            else:
                if kwargs.has_key('repeat'):
                    kwargs.pop('repeat')
                if kwargs.has_key('fps'):
                    kwargs.pop('fps')
                try:
                    kwargs.pop('ignore_anim')
                except:
                    pass
                texture = Texture(filename, **kwargs)

            self.__class__.textures[filepath] = texture

        return texture



    # Moved this from the renderable instances to here, lets us do some
    # optimisations regarding buffer binding
    def draw_vbo(self, vertex_number, vbo_vert, vbo_color, vbo_texcoord):
        if self.__class__.vbo_vert != vbo_vert: 
            glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_vert)
            glVertexPointer(2, GL_FLOAT, 0, None)
            self.__class__.vbo_vert = vbo_vert
        
        #if self.__class__.vbo_color != vbo_color:
            #glBindBuffer(GL_ARRAY_BUFFER, vbo_color)
            #glColorPointer(3, GL_FLOAT, 0, None)
            #self.__class__.vbo_color = vbo_color
        
        if self.__class__.vbo_texcoord != vbo_texcoord:
            glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_texcoord)
            glTexCoordPointer(2, GL_FLOAT, 0, None)
            self.__class__.vbo_texcoord = vbo_texcoord
        
        if not self.__class__.state_vert_array:
            glEnableClientState(GL_VERTEX_ARRAY)
            self.__class__.state_vert_array = True
            
        #if not self.__class__.state_color_array:
            #glEnableClientState(GL_COLOR_ARRAY)
            #self.__class__.state_color_array = True
            
        if not self.__class__.state_texcoord_array:
            glEnableClientState(GL_TEXTURE_COORD_ARRAY)
            self.__class__.state_texcoord_array = True
            
        glDrawArrays(GL_TRIANGLES, 0, vertex_number);
        
        
    def set_blend(self, type):
        """ sets the required blend function

        @param type: a blend type, may be 'replace', 'addictive', 'subtractive', 'modulate'
        """

        switch = {
            'replace': (0, 0),
            'addictive': (GL_ONE, GL_ONE),
            'subtractive': (GL_ZERO, GL_ONE_MINUS_SRC_COLOR),
            'modulate': (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA),
        }

        mode = switch[type]

        if type == 'replace':
            glDisable(GL_BLEND)
        else:
            glEnable(GL_BLEND)

        if self.__class__.blend_func == mode:
            return

        glBlendFunc(mode[0], mode[1])
        self.__class__.blend_func = mode
        
        
    #def create_shader(self, vert_filename, frag_filename):
        
        #shader = Shader()
        #shader.add_shader(GL_VERTEX_SHADER_ARB, data.filepath('programs', vert_filename))
        #shader.add_shader(GL_FRAGMENT_SHADER_ARB, data.filepath('programs', frag_filename))
        #shader.link_shaders()
        
        #return shader
        
    #def set_shader(self, shader):
        #if not shader:
            #glUseProgramObjectARB(0)
        #else:
            #glUseProgramObjectARB(shader.prog_id)


class Texture:

    """ we load texture data and such in here """
    #__slots__ = ['filename', 'id', 'repeats', 'size', 'aspect', 'filter_min', 'filter_max']
    

    def __init__(self, filename, size=(1, 1),
                 smooth=False, mipmap=True, repeats=('repeat', 'repeat')):


        print 'loading graphics file', filename
        #global target_surface 
        width, height = (None, None)
        self.repeats = (GL_REPEAT, GL_REPEAT)

        if repeats[0] == 'clamp':
            self.repeats = (GL_CLAMP, self.repeats[1])

        if repeats[1] == 'clamp':
            self.repeats = (self.repeats[0], GL_CLAMP)


        filepath = gamelib.data.filepath(filename)
        surf = IMG_Load(filepath)
        
        
        rgbsurf = SDL_ConvertSurface(surf, target_surface.format, 0)
        
        data = rgbsurf.pixels.as_ctypes()
        width = surf.w
        height = surf.h

        texture = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, texture)

        mipmap = False
        if not mipmap:
            glTexImage2D(GL_TEXTURE_2D, 
                        0, GL_RGBA, width, height, 0,
                        GL_RGBA, GL_UNSIGNED_BYTE, data )

        else:
            gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, width, height,
                            GL_RGBA, GL_UNSIGNED_BYTE, data)

        id = texture


        glBindTexture(GL_TEXTURE_2D, id)

        # here we now *definatly* have id, width, height

        self.id = id
        self.size = (width, height)
        self.aspect = self.size[1] / float(self.size[0])

        self.filter_min = GL_LINEAR
        self.filter_max = GL_LINEAR

        #if not smooth:
        self.filter_min, self.filter_max = (GL_NEAREST, GL_NEAREST)

        if mipmap:
            self.filter_min = GL_LINEAR_MIPMAP_LINEAR

        self.set_tex_params()
        
        #SDL_FreeSurface(rgbsurf)
        SDL_FreeSurface(surf)


    def set_tex_params(self):

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, self.filter_min)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, self.filter_max)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, self.repeats[0])
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, self.repeats[1])


class Animation():
    
    def __init__(self, filename, size=(1, 1), fps=11.235955056, repeat=True,
                 smooth=True, mipmap=True, repeats=('repeat', 'repeat')):
                     
        #for our animation we expect to be given a file from our animtion, 
        #we then basically just take all the png files in that directory as our
        #frames
        
        self.data_dir, self.basefile = os.path.split(gamelib.data.filepath(filename))
        self.fps = fps
        self.repeat = repeat
        self.finished_animation = False # this true if a non looping animation reaches its last frame
        self.started = False
        listfiles = os.listdir(self.data_dir)
        files = []
        for file in listfiles:
            if file[-3:] == 'png':
                files.append(file)
                
        
        
        def order_list(x, y):
            x = x.lower()
            y = y.lower()

            cmp = r'.*(?P<major>\d+)_(?P<minor>\d+).png'

            xinfo = re.match(cmp, x).groupdict()
            yinfo = re.match(cmp, y).groupdict()

            if int(xinfo['major']) != int(yinfo['major']):
                return int(xinfo['major']) - int(yinfo['major'])

            return int(xinfo['minor']) - int(yinfo['minor'])


        files.sort(order_list)
        
        self.textures = []
        rcontext = render_context()
        for file in files:
            file = os.path.join(self.data_dir, file)
            texture = rcontext.create_texture(file, smooth=smooth, mipmap=mipmap, repeats=repeats, ignore_anim=True)
            self.textures += [texture]
            
        self.last_time = SDL_GetTicks()
        self.current_time = SDL_GetTicks()
        self.frame_index = 0.0
        
    @propertyd
    def id():
        """ the position of our item """
        def fget(self): return self._get_current_id()
        def fset(self, value): pass
        
    def reset(self):
        self.started = False
        self.finished_animation = False
        
    def _get_current_id(self):
        
        if not self.started:
            self.last_time = SDL_GetTicks()
            self.frame_index = 0.0
            self.started = True
        
        current_time = SDL_GetTicks()
        frametime = 1000/self.fps
        frames_since_last_call = (current_time - self.last_time) / float(frametime)
        self.frame_index += frames_since_last_call
        if self.repeat:
            self.frame_index %= len(self.textures)
        else:
            self.frame_index = min(self.frame_index, len(self.textures) - 1) 
        self.last_time = current_time
        
        if self.frame_index == len(self.textures) -1:
            self.finished_animation = True
            
        return self.textures[int(math.floor(self.frame_index))].id



#class Shader(object):
    
    #vert_source = None
    #frag_source = None
    #vert_prog = None
    #frag_prog = None
    
    #shaders = []
    
    #def __init__(self):
        #self.required_extensions = [
                #"GL_ARB_fragment_shader",
                #"GL_ARB_vertex_shader",
                #"GL_ARB_shader_objects",
                #"GL_ARB_shading_language_100",
                #"GL_ARB_vertex_shader",
                #"GL_ARB_fragment_shader"]
                
        #self.check_extensions( self.required_extensions )
        
        #self.prog_id = glCreateProgramObjectARB()
        ##self.checkOpenGLError()
        #self.ready = False
        #self.enabled = False

    #def check_extensions( self, extensions ):
        #"""Check if all extensions in a list are present."""
        #for ext in extensions:
            #if ( not ext ):
                #print "Driver does not support ", ext
                #sys.exit()
                
    #def add_shader(self, shaderType, fileName):
        #"""Read a shader program from a file.

        #The program is load and compiled"""
        #global shader_support
        #shader_id = glCreateShaderObjectARB(shaderType)
        
        #source = open(fileName, 'r').read()
        #glShaderSourceARB(shader_id, [source])
        #glCompileShaderARB(shader_id)
        
        #success = glGetObjectParameterivARB(shader_id, GL_OBJECT_COMPILE_STATUS_ARB)
        #if (not success):
            #print glGetInfoLogARB(shader_id)
            #shader_support = False
            
        #glAttachObjectARB(self.prog_id, shader_id)
        #self.shaders += [shader_id]
        
    #def link_shaders(self):
        #"""Link compiled shader programs."""
        #global shader_support
        #glLinkProgramARB(self.prog_id)
        
        #success = glGetObjectParameterivARB(self.prog_id, GL_OBJECT_LINK_STATUS_ARB )
        #if (not success):
            #print glGetInfoLogARB(self.prog_id)
            #shader_support = False
        #else:
            #self.ready = True
            
        
    
class Camera(object):
    
    target_object = None
    cam_mod = (0.0, 0.0)
    current_position = (0.0, 0.0)
    xbounds = None
    rcontext = None
    
    def __init__(self, xbounds=None):
        if not self.__class__.rcontext:
            self.__class__.rcontext = render_context()
        
        if xbounds:
            self.__class__.xbounds = xbounds
        
        
    def track_object(self, object, mod=(0.0, 0.0)):
        self.__class__.target_object = object
        self.__class__.cam_mod = mod
    
    def move_to(self, position):
        self.__class__.target_position = position
        
    def set_position(self, position):
        self.__class__.current_position = position
        
    def set_screen(self):
        try:
            x, y = self.__class__.target_object.position
            x += self.cam_mod[0]
            y += self.cam_mod[1]
            self.__class__.current_position = (x, -y)
        except:
            pass
        
        #if self.__class__.target_object and hasattr(self.__class__.target_object, 'position'):
            #self.__class__.current_position = self.__class__.track_object.position
        
        if self.__class__.xbounds:
            x, y = self.__class__.current_position
            #x -= self.cam_mod[0]
            x -= 200
            x = max(x, self.__class__.xbounds[0])
            x += 400
            x = min(x, 1024*1024)#self.__class__.xbounds[1])
            x -= 200
            #x += self.cam_mod[0]
            self.__class__.current_position = (x, y)
        
        
        self.__class__.current_position = (self.__class__.current_position[0], -463)
        self.__class__.rcontext.set_screen_pos(*self.__class__.current_position)
        self.rcontext.set_screen()
        

#Okay these are the renderables, the idea here is that a lot of the 'shapes' 
# we draw are going to be quite simiar, basically quads and maybe some building
# shapes or something, so why re-bind things to buffers all the time
# thus we use the __class__ to make each child of this class have its own 
# vertex buffer object, but not each instance of each child

class renderable(object):
    """ renderables can be rendered to the screen 
        
    """
    _reference_count = 0
    _vertex_buffer = None
    _color_buffer = None
    _tex_buffer = None
    
    _vertex_array = []
    _num_verts = 0
    _color_array = []
    _tex_array = []
    
    _rcontext = render_context()
    
    textures = {}
    texture = None
    texture_scale = None
    color = None
    scale = None
    shader = None
    shader_mod = (0.0, 0.0)
    shader_mod_ticks = 0
    
    def __getattr__(self, attribute):
        if attribute == 'reference_count':
            return self._reference_count
        elif attribute == 'rcontext':
            return self._rcontext
        elif attribute == 'vertex_buffer':
            return self._vertex_buffer
        elif attribute == 'color_buffer':
            return self._color_buffer
        elif attribute == 'tex_buffer':
            return self._tex_buffer
        elif attribute == 'vertex_array':
            return self._vertex_array
        elif attribute == 'color_array':
            return self._color_array
        elif attribute == 'tex_array':
            return self._tex_array
        elif attribute == 'num_verts':
            return self._num_verts
        
        
    def __setattr__(self, attribute, value):
        if attribute == 'reference_count':
            self.__dict__['_reference_count'] = value
        elif attribute == 'rcontext':
            self.__dict__['_rcontext'] = value
        elif attribute == 'vertex_buffer':
            self.__dict__['_vertex_buffer'] = value
        elif attribute == 'color_buffer':
            self.__dict__['_color_buffer'] = value
        elif attribute == 'tex_buffer':
            self.__dict__['_tex_buffer'] = value
        elif attribute == 'vertex_array':
            self.__dict__['_vertex_array'] = value
        elif attribute == 'color_array':
            self.__dict__['_color_array'] = value
        elif attribute == 'tex_array':
            self.__dict__['_tex_array'] = value
        elif attribute == 'num_verts':
            self.__dict__['_num_verts'] = value
            
        else:
            self.__dict__[attribute] = value
    
    def __init__(self):
        self.texture = None
        self.textures = {}
        if self.reference_count <= 0:
            #our reference count is at 0 - thus we need to generate our various
            #generatables 
        
            #build our vertex buffer objects for this renderable
            # NOTE - this does not build a new vbo for each renderable instance
            # but rather it creates ONE for each subclass - much more efficent
            # - obviously less flexable, but screw it, make a new subclass if 
            # you want a different looking object
            self._gen_arrays()
            
        self.shader = None    
        
        self.reference_count += 1
        
    def __del__(self):
        self.reference_count -= 1
        
        if self.reference_count <= 0:
            # we lost all references to this class type, we can safely remove 
            # our pre-generated things
            
            glDeleteBuffers(self.vertex_buffer, 1)
            #glDeleteBuffers(self.color_buffer, 1)
            glDeleteBuffers(self.tex_buffer, 1)
            self.vertex_buffer = None
            self.color_buffer = None
            self.tex_buffer = None
            
    
    def add_texture(self, name, texture):
        self.textures[name] = texture
    
    def set_texture(self, name):
        self.texture = self.textures[name]
        
    def set_shader(self, shader):
        self.shader = shader
        self.shader_mod = (random.random(), random.random())

        
    def _gen_arrays(self):
        self.vertex_buffer = glGenLists(1)
        #self.color_buffer = glGenLists(1)
        self.tex_buffer = glGenLists(1)

        #upload the vertex array into the VBO
        glBindBufferARB(GL_ARRAY_BUFFER_ARB, self.vertex_buffer)
        glBufferDataARB(GL_ARRAY_BUFFER_ARB, numpy.array(self.vertex_array, dtype=numpy.float32), GL_STATIC_DRAW_ARB)
        
        
        ##upload the color array into the VBO
        #glBindBuffer(GL_ARRAY_BUFFER, self.color_buffer)
        #glBufferData(GL_ARRAY_BUFFER, numpy.array(self.color_array, dtype=numpy.float32), GL_STATIC_DRAW)
        
        
        #upload the texture array into the VBO
        glBindBufferARB(GL_ARRAY_BUFFER_ARB, self.tex_buffer)
        glBufferDataARB(GL_ARRAY_BUFFER_ARB, numpy.array(self.tex_array, dtype=numpy.float32), GL_STATIC_DRAW_ARB)
        
        #print(self.vertex_array, self.color_array, self.tex_array)
        
        
    def _draw_array(self):
        glBindBufferARB(GL_ARRAY_BUFFER_ARB, self.vertex_buffer)
        #print(self.vertex_buffer, self.color_buffer)
        glVertexPointer(2, GL_FLOAT, 0, None)
        
        #glBindBuffer(GL_ARRAY_BUFFER, self.color_buffer)
        #glColorPointer(3, GL_FLOAT, 0, None)
        
        glBindBufferARB(GL_ARRAY_BUFFER_ARB, self.tex_buffer)
        glTexCoordPointer(2, GL_FLOAT, 0, None)
        
        glEnableClientState(GL_VERTEX_ARRAY)
        #glEnableClientState(GL_COLOR_ARRAY)
        glEnableClientState(GL_TEXTURE_COORD_ARRAY)

        glDrawArrays(GL_TRIANGLES, 0, self.num_verts);
        
        glDisableClientState(GL_TEXTURE_COORD_ARRAY)
        #glDisableClientState(GL_COLOR_ARRAY)
        glDisableClientState(GL_VERTEX_ARRAY)
        
    def draw(self):
        if self.texture:
            # we do have a texture, so lets make sure to bind it first
            glEnable(GL_TEXTURE_2D)
            self.rcontext.bind_texture(self.texture)
        else:
            glDisable(GL_TEXTURE_2D)
            self.rcontext.unbind_textures()
        
        if self.color:
            glColor(*self.color)
        else:
            glColor(1.0, 1.0, 1.0)
            
        if self.scale:
            self.rcontext.scale_to(*self.scale)
        else:
            self.rcontext.scale_to(1.0, 1.0)
            
        if self.texture_scale:
            self.rcontext.texture_scale_to(*self.texture_scale)
        else:
            self.rcontext.texture_scale_to(1.0, 1.0)
            
        #if self.shader:
            #self.rcontext.set_shader(self.shader)
            #self.rcontext.bind_texture(self.texture, self.textures['shader_tex'])
             ## set our variables
            #loc = glGetUniformLocation(self.shader.prog_id, 'srcSampler')
            #glUniform1i(loc, 0)

            #loc = glGetUniformLocation(self.shader.prog_id, 'grainSampler')
            #glUniform1i(loc, 1)

            ##loc = glGetUniformLocation(self.shader.prog_id, 'brightness')
            ##glUniform1f(loc, 1.0)
            
            #loc = glGetUniformLocation(self.shader.prog_id, 'color')
            #color = self.color if len(self.color) == 4 else (self.color[0], self.color[1], self.color[2], 1.0)
            
            #glUniform4f(loc, *color)
                
            #loc = glGetUniformLocation(self.shader.prog_id, 'grain_mod')
            #glUniform2f(loc, *self.shader_mod)
            

        #else:
            #self.rcontext.set_shader(None)
            
            
        self.rcontext.draw_vbo(self.num_verts, self.vertex_buffer, self.color_buffer, self.tex_buffer)

        self.rcontext.scale_to(1.0, 1.0)


class renderable_static(renderable):
    """ this class is a modification of the renderable class, for classes which
    have many instances which share the same vbo/glprogram/whatever, it will
    share one glbuffer for all the instances of a given class
    """
    
    def __getattr__(self, attribute):
        if attribute == 'reference_count':
            return self.__class__._reference_count
        elif attribute == 'rcontext':
            return self.__class__._rcontext
        elif attribute == 'vertex_buffer':
            return self.__class__._vertex_buffer
        elif attribute == 'color_buffer':
            return self.__class__._color_buffer
        elif attribute == 'tex_buffer':
            return self.__class__._tex_buffer
        elif attribute == 'vertex_array':
            return self.__class__._vertex_array
        elif attribute == 'color_array':
            return self.__class__._color_array
        elif attribute == 'tex_array':
            return self.__class__._tex_array
        elif attribute == 'num_verts':
            return self.__class__._num_verts
        
        
    def __setattr__(self, attribute, value):
        if attribute == 'reference_count':
            self.__class__._reference_count = value
        elif attribute == 'rcontext':
            self.__class__._rcontext = value
        elif attribute == 'vertex_buffer':
            self.__class__._vertex_buffer = value
        elif attribute == 'color_buffer':
            self.__class__._color_buffer = value
        elif attribute == 'tex_buffer':
            self.__class__._tex_buffer = value
        elif attribute == 'vertex_array':
            self.__class__._vertex_array = value
        elif attribute == 'color_array':
            self.__class__._color_array = value
        elif attribute == 'tex_array':
            self.__class__._tex_array = value
        elif attribute == 'num_verts': 
            self.__class__._num_verts = value
        else:
            self.__dict__[attribute] = value
            
            
#as an example, here is a basic box type
class BasicBox(renderable_static):
    
    _vertex_array = [
        -0.5,-0.5,
         0.5,-0.5,
         0.5, 0.5,
         0.5, 0.5,
        -0.5, 0.5,
        -0.5,-0.5
    ]
    _num_verts = 6
    
    #_color_array = [
        #1.0, 0.0, 0.0,
        #1.0, 1.0, 0.0,
        #0.0, 1.0, 0.0,
        #0.0, 1.0, 0.0,
        #0.0, 0.0, 1.0,
        #1.0, 0.0, 0.0
    #]    _
    #color_array = [
        #1.0, 1.0, 1.0,
        #1.0, 1.0, 1.0,
        #1.0, 1.0, 1.0,
        #1.0, 1.0, 1.0,
        #1.0, 1.0, 1.0,
        #1.0, 1.0, 1.0
    #]
    
    _tex_array = [   
        0.0, 1.0,
        1.0, 1.0,
        1.0, 0.0,
        1.0, 0.0,
        0.0, 0.0,
        0.0, 1.0
    ]
