#!/usr/bin/env python
#
#       world.py
#       
#       Copyright 2009 Gordon Allott <gordallott@gmail.com>
#       
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 3 of the License, or
#       (at your option) any later version.
#       
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#       
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
from gamelib import render, scenegraph, svg_parse, collision, input, audio
from gamelib.ai import base as ai
import os, sys
import random


entity_definitions = {
    'jason': 
    {
        'ai': ai.Jason,
        'textures':
        {
            'walk_anim': ('graphics/anim/jason_walk/jason_walk_0_0.png', {'fps':11.235955056}),
            'walk_strut': ('graphics/anim/jason_strut/jason-strut_0_0.png', {'fps':11.235955056}),
            'walk_shuffle': ('graphics/anim/jason_shuffle/jason_shuffle_0_0.png', {'fps':11.235955056}),
            'drink_beer': ('graphics/anim/drink_beer/jason_drink_beer_0_4.png', {'repeat':False}),
            'drink_coffee': ('graphics/anim/drink_coffee/jason_drink_coffee_0_4.png', {'repeat':False}),
            'guitar_take': ('graphics/anim/jason_guitar/ajason_guitar_take_0_0.png', {'repeat': False, 'fps':7}),
            'proposal': ('graphics/anim/jason_proposing/jason_proposing_0_0.png', {'repeat':False}),
            'work': ('graphics/anim/jason_work/jason_work_0_0.png', {'repeat': False}),
            'angry_start': ('graphics/anim/jason_angry_start/jason_angry_start_0_0.png', {'repeat': False}),
            'angry': ('graphics/anim/jason_angry/jason_angry_0_0.png', {'repeat':True}),
            'sitting': ('graphics/anim/jason_sitting/jason_sitting_0_0.png', {'repeat':False}),
            'drink_juice': ('graphics/anim/drink_juice/jason_drink_juice_0_0.png', {'repeat':False}),
            'drink_wine': ('graphics/anim/drink_wine/jason_drink_wine_0_0.png', {'repeat': False}),
        },
        #'mouse_sensitive': True,
        'shape':[(-64.0, -64.0), (64.0, -64.0), (64.0, 64.0), (-64.0, 64.0)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    'spouse_courting':
    {
        'ai': ai.spouse_courting,
        'textures':
        {
            'walk_anim': ('graphics/anim/girl_walk/girl_walk_0_0.png', {}),
            'stand': ('graphics/anim/girl/girl_0_0.png', {}),
            'proposal': ('graphics/anim/girl_proposal/girl_proposal_0_0.png', {'repeat':False}),
        },
        'mouse_sensitive': True,
        'shape':[(-64.0, -64.0), (64.0, -64.0), (64.0, 64.0), (-64.0, 64.0)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    
    
    'spouse_sitting':
    {
        'ai': ai.spouse_sitting,
        'textures':
        {  
            'default': ('graphics/anim/girl_sit/girl_sit_0_0.png', {}),
            'withflower': ('graphics/girl_sitflower_0_0.png', {}),
        },
        'mouse_sensitive': True,
        'shape': [(-22.0, -50.0), (2.0, -50.0), (22.0, 50.0), (-22.0, 50.0)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    
    'spouse_sleeping':
    {
        'ai': ai.spouse_sleeping,
        'textures':
        {
            'default': ('graphics/house/house_bedroom_wife.png', {}),
        },
        'mouse_sensitive': False,
        'shape': [(-0.0, -0.0), (0.0, -0.0), (0.0, 0.0), (-0.0, 0.0)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    
    'spouse_endsit':
    {
        'ai': ai.spouse_endsit,
        'textures':
        {  
            'default': ('graphics/anim/girl_sit/girl_sit_0_0.png', {}),
        },
        'mouse_sensitive': False,
        'shape': [(-22.0, -50.0), (2.0, -50.0), (22.0, 50.0), (-22.0, 50.0)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    
    'child_play':
    {
        'ai': ai.child_play,
        'textures':
        {
            'default': ('graphics/anim/child_play/child_play_0_0000.png', {'fps':3}),
        },
        'mouse_sensitive': True,
        'shape': [(-72.0, -40.0), (72.0, -40.0), (72.0, 40.0), (-72.0, 40.0)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    
    'child_chair':
    {
        'ai': ai.child_chair,
        'textures':
        {
            'default': ('graphics/anim/child/child_homework_0_0.png', {'fps':3}),
        },
        'mouse_sensitive': True,
        'shape': [(-38.0, -32.0), (38.0, -32.0), (38.0, 32.0), (-38.0, 32.0)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    
    'child_end':
    {
    'ai': ai.child_end,
        'textures':
        {
            'default': ('graphics/anim/child_play/child_play_0_0000.png', {'fps':3}),
        },
        'mouse_sensitive': False,
        'shape': [(-72.0, -40.0), (72.0, -40.0), (72.0, 40.0), (-72.0, 40.0)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    
    'ring':
    {
        'ai': ai.ring,
        'textures':
            {'default': ('graphics/anim/engagement_ring/engagement_ring_0_0.png', {})},
        'mouse_sensitive': True,
        'shape': [(-11.0, -8.0), (11.0, -8.0), (11.0, 8.0), (-11.0, 8.0)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    
    'flower':
    {
        'ai': ai.flower,
        'textures':
            {'default': ('graphics/house/house_flower.png', {})},
        'mouse_sensitive': True,
        'shape': [(-11.0, -8.0), (11.0, -8.0), (11.0, 8.0), (-11.0, 8.0)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },


    'juice':
    {
        'ai': ai.juice,
        'textures':
        {
            'default': ('graphics/house/house_juice.png', {}),
        },
        'mouse_sensitive': True,
        'shape': [(-6, -8), (6, -8), (6, 8), (-6, 8)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    
    'wine':
    {
        'ai': ai.wine,
        'textures':
        {
            'default': ('graphics/house/house_wine.png', {}),
        },
        'mouse_sensitive': True,
        'shape': [(-6, -12), (6, -12), (6, 12), (-6, 12)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    
    'soap':
    {
        'ai': ai.soap,
        'textures':
        {
            'default': ('graphics/house/house_soap.png', {}),
        },
        'mouse_sensitive': True,
        'shape': [(-6, -8), (6, -8), (6, 8), (-6, 8)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    
    'tv':
    {
        'ai': ai.tv,
        'textures':
        {
            'default': ('graphics/house/tv_off_0_0.png', {}),
            'on': ('graphics/anim/tv/tv_on_0_0.png', {'repeat':True}),
        },
        'mouse_sensitive': True,
        'shape': [(-36, -46), (36, -46), (36, 46), (-36, 46)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    
    'toothbrush':
    {
        'ai': ai.toothbrush,
        'textures':
        {
            'default': ('graphics/house/house_toothbrush.png', {}),
        },
        'mouse_sensitive': True,
        'shape': [(-12, -6), (12, -6), (12, 6), (-12, 6)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },

    'baby':
    {
        'ai': ai.baby,
        'textures':
            {'default': ('graphics/house/house_babby.png', {})},
        'mouse_sensitive': True,
        'shape': [(-64.0, -32.0), (64.0, -32.0), (64.0, 32.0), (-64.0, 32.0)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },

    #--- Things for jason to consume or whatever ---
    'sky':
    {
        'ai': ai.sky,
        'textures':
            {
            'afternoon': ('graphics/tileable/afternoon_sky.png', {}),
            'evening': ('graphics/tileable/evening_sky.png', {}),
            'morning': ('graphics/tileable/morning_sky.png', {}),
            'night': ('graphics/tileable/night_sky.png', {})
            },
        'mouse_sensitive': False,
        'shape': [(-0.0, -0.0), (0.0, -0.0), (0.0, 0.0), (-0.0, 0.0)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    
    'pub_beer':
    {
        'ai': ai.pub_beer,
        'textures':
        {
            'default': ('graphics/pub/pub_beer.png', {}),
        },
        'mouse_sensitive': True,
        'shape': [(-14, -14), (14, -14), (14, 14), (-14, 14)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    'pub_guitar':
    {
        'ai': ai.pub_guitar,
        'textures':
        {
            'default': ('graphics/pub/pub_guitar.png', {}),
        },
        'mouse_sensitive': True,
        'shape': [(-14, -38), (14, -38), (14, 38), (-14, 38)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    
    'office_drone':
    {
        'ai': ai.office_drone,
        'textures':
        {
            'default': ('graphics/anim/worker_drone/worker_drone_0_0.png', {'repeat':True}),
            'give_paper': ('graphics/anim/worker_work/worker_work_0_0.png', {'repeat':False}),
        },
        'mouse_sensitive': True,
        'shape': [(-54, -50), (54, -50), (54, 50), (-54, 50)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    'coffee_pot':
    {
        'ai': ai.coffee_pot,
        'textures':
        {
            'default': ('graphics/office/office_coffeepot.png', {}),
        },
        'mouse_sensitive': True,
        'shape': [(-18, -14), (18, -14), (18, 14), (-18, 14)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    
    
    'credit_result':
    {
        'ai': ai.credit_result,
        'textures':
        {
            'a': ('graphics/credits/endinga.png', {}),
            'b': ('graphics/credits/endinga.png', {}),
            'c': ('graphics/credits/endinga.png', {}),
            'd': ('graphics/credits/endinga.png', {}),
        },
        'mouse_sensitive': False,
        'shape': [(0, 0), (0, -0), (0, 0), (-0, 0)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    
    'credits':
    {
        'ai': ai.credits,
        'textures':
        {
            'default': ('graphics/credits/ending.png', {}),
        },
        'mouse_sensitive': False,
        'shape': [(0, 0), (0, -0), (0, 0), (-0, 0)],
        #'shaders': ('basic.vert.glsl', 'test.frag.glsl'),
        #'shader_tex': ('graphics/overlay_paper.png'),
    },
    
    'test_hitbox':
    {
        'ai': ai.AI, #generic AI, does nothing really
        'norender': True,
        'shape': [(-24.0, -48.0), (24.0, -48.0), (24.0, 48.0), (-24.0, 48.0)],
    },
    
    
    # -- things that arn't really there, that just do stuff ---
    'act_change':
        {
        'ai': ai.act_changer,
        'norender':True,
        'shape': [(-24.0, -48.0), (24.0, -48.0), (24.0, 48.0), (-24.0, 48.0)],
    },
        
        
    'end_start':
    {
        'ai': ai.end_start,
        'norender': True,
        'shape': [(64, -48.0), (64, -48.0), (64, 48.0), (64, 48.0)],
    },
    
    #Sound handling AI's
    'sound_bar':
    {
        'ai': ai.bar_sound,
        'norender': True,
        'shape': [(-1, -200), (1, -200), (1, 200), (-1, 200)],
    },
    'sound_home':
    {
        'ai': ai.home_sound,
        'norender': True,
        'shape': [(-1, -200), (1, -200), (1, 200), (-1, 200)],
    },
    'sound_office':
    {
        'ai': ai.office_sound,
        'norender': True,
        'shape': [(-1, -200), (1, -200), (1, 200), (-1, 200)],
    },
    'sound_schoolbell':
    {
        'ai': ai.schoolbell,
        'norender': True,
        'shape': [(-1, -200), (1, -200), (1, 200), (-1, 200)],
    },
    
    
    
    
}

# Found this on the internet somewhere, active state cookbook maybe?
# it might of been a wikipage i found for python decorators
# basically it just lets me set the set_attr and get_attr for variables quickly
def propertyd(function):
    keys = 'fget', 'fset', 'fdel'
    func_locals = {'doc':function.__doc__}
    def probe_func(frame, event, arg):
        if event == 'return':
            locals = frame.f_locals
            func_locals.update(dict((k, locals.get(k)) for k in keys))
            sys.settrace(None)
        return probe_func
    sys.settrace(probe_func)
    function()
    return property(**func_locals)

class Entity(object):
    """ entities exist to live inside a quadtree """
    
    def __init__(self, type='generic', x=0.0, y=0.0, layer=0, id=None):
        self.qnode = None
        self._position = (x, y)
        self._rotation = 0.0
        self._scale = (1.0, 1.0)
        self.type = type
        self.id = id
        
        self.renderable = None
        self.ai = None
        self.shape = None
        self.set_layer(layer)
        self.mouse_sensitive = False
        

        
        #we 'jitter' the zposition of the entity, this is so we don't have 
        # lots of objects z-fighting all day
    
    def set_layer(self, layer):
        self.layer = layer
        self.zpos = (random.random()*0.95) - 0.5
        self.cached_z = self.layer + self.zpos
    
    def update_qnode(self):
        if self.qnode:
            self.qnode.move_item(self)
    
    def set_type(self, type):
        self.type = type 
        
    @propertyd
    def position():
        """ the position of our item """
        def fget(self): return self._position
        def fset(self, newpos): 
            self._position = newpos
            self.update_qnode()
            if self.shape:
                self.shape.set_position(self.position)
            
    @propertyd
    def x():
        """ the x position """
        def fget(self): return self.position[0]
        def fset(self, newx): self.position = (newx, self.position[1])
    
    @propertyd
    def y():
        """ the y position """
        def fget(self): return self.position[1]
        def fset(self, newy): self.position = (self.position[0], newy)

    @propertyd
    def rotation():
        """ the rotation of the entity """
        def fget(self): return self._rotation
        def fset(self, newrot): self._rotation = newrot
    
    @propertyd
    def scale():
        """ the scale of the entiry """
        def fget(self): return self._scale
        def fset(self, newscale): self._scale = newscale

    def set_renderable(self, renderable, *args, **kwargs):
        """ sets the given renderable as this entities graphic """
        if callable(renderable):
            self.renderable = renderable(*args, **kwargs)
        else:
            self.renderable = renderable
            
    def set_ai(self, ai_system, *args, **kwargs):
        """ sets the given ai_system as this entities ai """
        if callable(ai_system):
            self.ai = ai_system(self, *args, **kwargs)
        else:
            self.ai = ai_system
            
    def set_shape(self, shape):
        self.shape = shape
         
        
    def draw(self):
        ''' draws the renderable '''
        if self.renderable:
            self.renderable.draw() 
        
    def tick(self):
        self.ai.tick()
        
    def destroy(self):
        # Careful here, need to at some point make sure that per_tick entitys 
        # get handled nicely
        
        del(self.renderable)
        self.renderable = None
        del(self.ai)
        self.ai = None
        del(self.shape)
        self.shape = None
        
        #this is sort of annoying, this does not work for some reason
        #self.qnode is not acctually (as far as i can tell) pointing to the
        #right qnode and thus remove_item does not work. 
        #self.qnode.remove_item(self)
        
    def __str__(self):
        return "Entity object (%s - %s) at %i, %i" % (self.id, self.type, self.x, self.y)
        

import math

class Pointer(object):
    """ we handle the pointer in here """
    
    input = None
    camera = None
    physworld = None
    glow = 0.0
    last_entity = None
    stats = None
    
    def __init__(self):
        if not self.__class__.input:
            self.__class__.input = input.Input()
        if not self.__class__.camera:
            self.__class__.camera = render.Camera()
        if not self.__class__.physworld:
            self.__class__.physworld = collision.physworld()
            
        if not self.__class__.stats:
            self.__class__.stats = ai.Statistics()
            
        self.__class__.x = 0.0
        self.__class__.y = 0.0
            
    def tick(self):
        mx, my = self.__class__.input.mouse_x, self.__class__.input.mouse_y
        mx = 400 * (mx - 0.5)
        my = 300 * (my - 0.5)
        self.__class__.x = self.__class__.camera.current_position[0] + mx 
        self.__class__.y = (-self.__class__.camera.current_position[1]) - my
        
        #make the glow happen
        self.__class__.glow += 1.0/10.0
        #self.__class__.glow %= 1
        
        self.__class__.physworld.mouse_update((self.__class__.x, self.__class__.y))
                
        if self.__class__.last_entity:
            if self.__class__.last_entity.renderable:
                self.__class__.last_entity.renderable.color = (1.0, 1.0, 1.0)
        
        if self.__class__.last_entity != self.__class__.physworld.mouse_entity:
            self.__class__.glow = 1.0
            
        self.__class__.last_entity = self.__class__.physworld.mouse_entity
        if self.__class__.last_entity:
            if self.__class__.last_entity.renderable != None:
                glow_percentage = 1.0 - (((math.cos(self.__class__.glow) * 0.5) + 0.5) * 0.5)
                self.__class__.last_entity.renderable.color = (glow_percentage, glow_percentage, glow_percentage)
             
        if self.input.mouse_down:
            
            if hasattr(self.__class__.physworld.mouse_entity, 'triggered'):
                # if we have a triggered attribute, check its not already fired
                if not self.__class__.physworld.mouse_entity.triggered:
                    self.__class__.stats.attention = self.__class__.last_entity
            else:
                self.__class__.stats.attention = self.__class__.last_entity
            
            
        
        #print self.__class__.physworld.mouse_entity
        
    def get_position(self):
        return (self.__class__.x, self.__class__.y)
        

class World(object):
    
    def __init__(self, screen_size=None, world_name='world'):
        self.end_act = False
        self.rcontext = render.render_context()
        self.input = input.Input()
        self.aimanager = ai.AI_manager()
        self.aimanager.reset()
        self.x_pos = 0.0
        self.y_pos = 0.0
        self.camera = render.Camera(xbounds=(0.0, 1024.0*10))
        self.pointer = Pointer()
        self.world_name = world_name
        
        self.statistics = ai.Statistics()
        self.statistics.time = 200
        self.statistics.__class__.is_married
        
        self.ai_endact = ai.act_changer(None)
        self.ai_endact.add_callback(self.end_act_callback)
        
        self.per_frame_ticks = []
        
        
        # we create our scenegraph as a normal node, the bounds we specify 
        # basically have to be the upper limit of our world size, so make sure we 
        # get it right, smaller bounds = more efficent
        self.scenegraph = scenegraph.Node(bounds=scenegraph.Rect(0.0, 0.0, 1024.0*100, 1024*3.0))
        self.physworld = collision.physworld()
        #self.shape1 = collision.Shape('shape1')
        #self.shape2 = collision.Shape('shape2')
        
        #self.shape1.new_body(position=(500.0, 277.0))
        ##self.shape1.new_shape_circle(200.0)
        #self.shape1.new_shape_poly([(-200, -200), (200, -200), (200, 200), (-200, 200)])
        
        #self.shape2.new_body(position=(-1.0, 0.0))
        #self.shape2.new_shape_circle(5.0)
        
        #self.physworld.add(self.shape1)
        #self.physworld.add(self.shape2)
        
        texture = self.rcontext.create_texture('test.png')
        
        layers = svg_parse.parseXML("%s.svg" % (self.world_name))
        for layer in layers.iteritems():
            layer = layer[1]
            #print 'creating layer', layer['id']
            
            for item in layer['items']:
                #print 'creating item', item['id']
                entity = Entity(layer=layer['z'], id=item['id'])
                entity.position = (item['x'], item['y'])
                entity_def = {} 
                
                if item['entdef']:
                    if not entity_definitions.has_key(item['entdef']):
                        print 'Entity definition', item['entdef'], 'not found'
                    else:
                        print 'Found Entity', item['entdef']
                        entity_def = entity_definitions[item['entdef']]
                        entity.set_type(item['entdef'])
                
                if entity_def.has_key('mouse_sensitive'):
                    entity.mouse_sensitive = entity_def['mouse_sensitive']
                
                if item['type'] == 'rect':
                    #print 'creating rect'
                    
                    
                    if not entity_def.has_key('norender'):
                        renderable = render.BasicBox()
                        renderable.color = item['style']['color']
                        renderable.scale = (item['width'], item['height'])
                        entity.set_renderable(renderable)
                        
                        if item['texture']:
                            repeat = False
                            if item['texture_tile']:
                                repeat = True
                            
                            
                                
                            texture = self.rcontext.create_texture(item['texture'])#, repeats=repeat)
                            renderable.add_texture(item['texture'], texture)
                            renderable.set_texture(item['texture'])
                
                        if entity_def.has_key('textures'):
                            #print 'loading extra textures'
                            for name, data in entity_def['textures'].iteritems():
                                filename, kwargs = data 
                                #print data, filename, kwargs
                                print 'loading texture', name
                                texture = self.rcontext.create_texture(filename, **kwargs)
                                renderable.add_texture(name, texture)
                            #print 'done loading extra textures'
                    
                        #if entity_def.has_key('shaders'):
                            #if entity_def['shaders']:
                                #shader = self.rcontext.create_shader(*entity_def['shaders'])
                                #renderable.set_shader(shader)
                            
                            #if entity_def['shader_tex']:
                                #texture = self.rcontext.create_texture(entity_def['shader_tex'])#, repeats=repeat)
                                #renderable.add_texture('shader_tex', texture)
                                #renderable.set_texture('shader_tex')

                        
                    if entity_def:
                        print 'setting ai!', entity_def['ai']
                        entity.set_ai(entity_def['ai'], data=item['entdata'])
                        if entity.ai.call_on_tick:
                            self.per_frame_ticks += [entity]
                            
                        shape = collision.Shape(entity)
                        shape.new_body(position=entity.position)
                        shape.new_shape_poly(entity_def['shape'])
                        self.physworld.add(shape)
                        
                        entity.set_shape(shape)
                    
                    
                    
                
                    
                    
                    
                            
                                
                            
                            
                    
                    self.scenegraph.add_item(entity)
                    
        
        # we need to create a special entity for our screen-space collisions
        self.screen_space = Entity(type='screen_space')
        self.screen_space.set_ai(ai.ScreenSpace)
        screen_shape = collision.Shape(self.screen_space)
        screen_shape.new_body(position=(0.0, 0.0))
        #create a shape the size of our screen
        screen_shape.new_shape_poly([(-300.0, -240.0), (300.0, -240.0), (300.0, 240.0), (-300.0, 240.0)])
        self.physworld.add(screen_shape)
        self.screen_space.set_shape(screen_shape)
        #call every frame to update
        self.per_frame_ticks += [self.screen_space]
        
        #self.protagonist = Entity()
        #pro_ai = ai.Protagonist(parent_entity=self.protagonist)
        #self.scenegraph.add_item(
        #self.per_frame_ticks += [self.protagonist]
        
        #self.acontext = audio.audio_context()
        #self.test_sample = audio.audio_sample('john_Sauter__Electric_school_bell.ogg')
        #self.test_sample.play_sample()
        
        #self.pointer_entity = Entity(layer=2000, type='pointer_icon', id='pointer_icon')
        ##this entity basically just exists as a debug helper
        #self.pointer_entity.set_renderable(render.BasicBox)
        
        #texture = self.rcontext.create_texture('graphics/pub/pub_beer.png')
        #self.pointer_entity.renderable.add_texture('pointer', texture)
        #self.pointer_entity.renderable.set_texture('pointer')
        #self.pointer_entity.renderable.scale = (28, 28)
        #self.scenegraph.add_item(self.pointer_entity)
        #px, py = self.pointer.get_position()
        #py = -py
        #self.pointer_entity.position = (px, py)
        
        
                    
        
    def tick_world(self):
        """ moves the world forward one tick """
        for item in self.per_frame_ticks:
            item.tick()
        
        self.pointer.tick()
        self.aimanager.do_tick()  
        #self.pointer_entity.position = self.pointer.get_position()
        self.physworld.tick()
        
        self.statistics.tick()
        
        #self.test_sample.set_distance_angle(self.test_sample.distance + 0.01, self.test_sample.angle - 0.01) 
        
            
    def draw_world(self):
        """ draws the current world state to the screen """
        
        def zmap(a, b):
            if a.cached_z > b.cached_z:
                return 1
            elif a.cached_z == b.cached_z:
                return 0
            else:
                return -1 
            #value = int((a.layer + a.zpos) - (b.layer + b.zpos))
            #print value
            #return value
        
        self.camera.set_screen()
        
        px, py = self.camera.current_position
        py = -py
        screen_bounds = scenegraph.Rect(px, py, 400*5.0, 225.0*3.0)

        items = list(self.scenegraph.hit(screen_bounds))
        
        if self.input.do_left:
            self.x_pos -= 10.0
        elif self.input.do_right:
            self.x_pos += 10.0
        
        if self.input.do_down:
            self.y_pos += 10.0
        elif self.input.do_up:
            self.y_pos -= 10.0
            
        #self.camera.set_position((self.x_pos, self.y_pos))
        
        
        #self.rcontext.set_screen_pos(1024/2, 0.0)
        #self.rcontext.set_screen()
        
        items.sort(zmap)
        for item in items:
            if item.type == 'jason':
                self.statistics.draw()
                
            self.rcontext.move_to(*item.position)
            item.draw()
            
        #base_pos = self.camera.current_position
        #base_pos = (base_pos[0] - self.camera.cam_mod[0], (-base_pos[1] + 32))
        

    def end_act_callback(self):
        self.end_act = True

    def end(self):
        print 'ending!'
        screen_bounds = scenegraph.Rect(0, 0, 10240000000000, 10240000000000)
        items = list(self.scenegraph.hit(screen_bounds))
        for item in items:
            item.destroy()



class menu(object):
    
    
    def __init__(self, *args, **kwargs):
    
        self.rcontext = render.render_context()
        
        renderable = render.BasicBox()
        texture = self.rcontext.create_texture('graphics/titles/menu.png')#, repeats=repeat)
        renderable.add_texture('default', texture)
        renderable.set_texture('default')
        renderable.scale = (400, 225)
        self.menu_screen = renderable
        
        renderable = render.BasicBox()
        texture = self.rcontext.create_texture('graphics/titles/mbutton.png')#, repeats=repeat)
        renderable.add_texture('default', texture)
        renderable.set_texture('default')
        renderable.scale = (45.0, 20)
        self.button = renderable
        
        self.input = input.Input()
        self.end_act = False
        
    def tick_world(self, *args, **kwargs):
        if self.input.button_down or self.input.mouse_down:
            self.end_act = True
        
    def draw_world(self, *args, **kwargs):
        self.rcontext.move_to(0.0, 0.0)
        self.menu_screen.draw()
        self.rcontext.move_to(0.0, -100.0)
        self.button.draw()
        
    def end(self, *args, **kwargs):
        pass
